/*
  Charlieplexing.cpp - Using MsTimer2() with 1ms resolution
  
  Alex Wenger <a.wenger@gmx.de> http://arduinobuch.wordpress.com/
  Matt Mets <mahto@cibomahto.com> http://cibomahto.com/
  
  Timer init code from MsTimer2 - Javier Valencia <javiervalencia80@gmail.com>
  Misc functions from Benjamin Sonnatg <benjamin@sonntag.fr>
  
  History:
    2009-12-30 - V0.0 wrote the first version at 26C3/Berlin
    2010-01-01 - V0.1 adding misc utility functions 
      (Clear, Vertical,  Horizontal) comment are Doxygen complaints now
    2010-05-27 - V0.2 add double-buffer mode
    
    2010-11-24
      Modified for ikkeiplexing shield (direct two 8x8 dot matrix LED shield)
      With brightness equalization
	  by ikkei <ikkei@zeus.eonet.ne.jp> http://blog.goo.ne.jp/jh3kxm
	2012-12-23
	  Modified for Titchaino with switch
      by ikkei

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

//#include "WProgram.h" // for Arduino 022-
#include <arduino.h> // for Arduino 1.0
//#include <inttypes.h>
#include <avr/interrupt.h>
#include "Charlieplexing.h"
#include <MsTimer2.h>

//#define ikkeiplexing // ikkeiplexing shield connections
//#define titchaino // ikkei's 8x16duino (Titchaino and Chookleino) connections
//#define OnBoard // for OnBoard breadboard
//#define Dotsduino11 // Dotsduino by arms22
//#define Dotsduino12d // Dotsduino by arms22
#define SQUARE8 // 8 SQUARE

#ifdef ikkeiplexing
#define LK1 1
#define LK2 4
#define LK3 A1
#define LK4 7
#define LK5 A2
#define LK6 8
#define LK7 11
#define LK8 12
#define LA1 6
#define LA2 10
#define LA3 A3
#define LA4 2
#define LA5 A0
#define LA6 3
#define LA7 9
#define LA8 5
#endif

#ifdef titchaino
#define LK1 A3
#define LK2 13
#define LK3 9
#define LK4 11
#define LK5 3
#define LK6 A1
#define LK7 4
#define LK8 7
#define LA1 12
#define LA2 5
#define LA3 6
#define LA4 A2
#define LA5 8
#define LA6 10
#define LA7 A0
#define LA8 2
#endif

#ifdef OnBoard
#define LK1 A3
#define LK2 12
#define LK3 13
#define LK4 A0
#define LK5 5
#define LK6 7
#define LK7 3
#define LK8 2
#define LA1 9
#define LA2 4
#define LA3 8
#define LA4 6
#define LA5 A2
#define LA6 A1
#define LA7 11
#define LA8 10
#endif


#ifdef Dotsduino11 // Dotsduino connections
#define LK1 9
#define LK2 4
#define LK3 10
#define LK4 6
#define LK5 A3
#define LK6 11
#define LK7 A2
#define LK8 13
#define LA1 5
#define LA2 A1
#define LA3 A0
#define LA4 8
#define LA5 12
#define LA6 7
#define LA7 3
#define LA8 2
#endif

#ifdef Dotsduino12d // Dotsduino connections
#define LK1 2
#define LK2 3
#define LK3 4
#define LK4 13
#define LK5 A0
#define LK6 A1
#define LK7 A2
#define LK8 A3
#define LA1 12
#define LA2 11
#define LA3 10
#define LA4 9
#define LA5 5
#define LA6 6
#define LA7 7
#define LA8 8
#endif

#ifdef SQUARE8 // 8 SQUARE
#define LK1 A1
#define LK2 2
#define LK3 13
#define LK4 A4
#define LK5 5
#define LK6 12
#define LK7 6
#define LK8 9
#define LA1 A5
#define LA2 7
#define LA3 8
#define LA4 A2
#define LA5 11
#define LA6 A3
#define LA7 3
#define LA8 4
#endif

//Charlieplexing iLedSign;
//void (*Charlieplexing::func)();

/* -----------------------------------------------------------------  */
/** Table for the LED multiplexing cycles, containing 16 cycles made out of three bytes
 */
byte leds[2][48];

/// Determines whether the display is in single or double buffer mode
byte displayMode;

/// Flag indicating that the display page should be flipped as soon as the
/// current frame is displayed
boolean videoFlipPage;

/// Pointer to the buffer that is currently being displayed
byte* displayBuffer;

/// Pointer to the buffer that should currently be drawn to
byte* workBuffer;

//boolian frame = false; //frame, 10ms, 100ms



const int sw1 = LA5;
const int sw2 = LA6;
const int sw3 = LA7;
const int sw4 = LA8;
const int sw5 = LA5;
const int sw6 = LA6;
const int sw7 = LA7;
const int sw8 = LA8;
#ifdef SQUARE8
const int sw9 = A0;
const int sw10 = A0;
#else
const int sw9 = A4;
const int sw10 = A5;
#endif


static byte in_d = 0;
static byte in_b = 0;
static byte in_c = 0;
static byte out_d = 0;
static byte out_b = 0;
static byte out_c = 0;
static byte sw1234;
static byte swdata;
static byte period;
static byte current_sw;
static byte previous_sw;
static byte scan;// 16 Cycles of Matrix
static byte msec;

//###
#define MID_LED 5
#define MAX_LED 8
#define FRAME 0x01
#define TENMS 0x02

//### brightness equalization parameter
const int dd[MAX_LED] = {1, 300, 350, 420, 520, 650, 800, 950 };	//### duty ratio

/* -----------------------------------------------------------------  */
/** Table for LED Position in leds[] ram table 
 */

byte ledMap[16] = {
	LK1,LK2,LK3,LK4,LK5,LK6,LK7,LK8,LA1,LA2,LA3,LA4,LA5,LA6,LA7,LA8
	};

// all output Hi-Z
void alloff( void ){
	DDRD &= ~out_d; // line 76543210
	PORTD &= ~out_d;
	DDRB &= ~out_b; // line xxDCBA98
	PORTB &= ~out_b;
	DDRC &= ~out_c; // line xxJIHGFE
	PORTC &= ~out_c;
	DDRC &= ~0x30; // A4,A5
}

/* -----------------------------------------------------------------  */
/** The Interrupt code goes here !  
 */
void scanLED(){
	alloff(); // all output high-Z (input)

	if ((scan==8)||(scan==9)){ // switch common lines
		PORTD |= in_d;	// pull up register of input port
		PORTB |= in_b;	// pull up register of input port
		PORTC |= in_c;	// pull up register of input port
	}

	byte n = ledMap[ scan ]; // scan column output LOW for cathode
	if (n < 8) {
		DDRD  = _BV(n);
	} else if (n < 14 ){
		DDRB  = _BV(n-8);
	} else {
		DDRC  = _BV(n-14);
	} 

	//### count ON LED numbers
	byte on_count = 0;
	for (int j=0; j<8; j++){
		if (displayBuffer[scan*3] & 1<<j ){
		    on_count++;
		}
	}
	for (int j=0; j<6; j++){
		if (displayBuffer[scan*3+1] & 1<<j){
		    on_count++;
		}
		if (displayBuffer[scan*3+2] & 1<<j){
		    on_count++;
		}
	}

	if (scan==8){ // get SW1,SW2,SW3,SW4
		sw1234 = (digitalRead(sw1)*8 + digitalRead(sw2)*4 + digitalRead(sw3)*2 + digitalRead(sw4)) ^ 0x0F;
	} else if (scan==9){ // get SW5,SW6,SW7,SW8
		previous_sw = current_sw;
#ifdef ikkeiplexing
		current_sw = (sw1234 << 4) | ((digitalRead(sw5)*8 + digitalRead(sw6)*4 + digitalRead(sw7)*2 + digitalRead(sw8)) ^ 0x0F);
#else
		current_sw = ((digitalRead(sw9)<<1 | digitalRead(sw10)) ^ 0x03) << 6;
#endif
		if (previous_sw == current_sw){
			swdata = current_sw;
		}
	} 
	
	// swdata: SW1,SW2,SW3,SW4,SW5,SW6,SW7,SW8
	if ((scan==8)||(scan==9)){ // switch common lines
		PORTD &= ~in_d;	// pull up register of input port
		PORTB &= ~in_b;	// pull up register of input port
		PORTC &= ~in_c;	// pull up register of input port
	}

	//### equalization pre-delay
	if ( (on_count >= MID_LED) && (on_count < MAX_LED) ){
		delayMicroseconds( 1000 - dd[on_count] );
	}

	// output row data
	DDRD  |= displayBuffer[scan*3];
	PORTD  = displayBuffer[scan*3];
	DDRB  |= displayBuffer[scan*3+1];
	PORTB  = displayBuffer[scan*3+1];
	DDRC  |= displayBuffer[scan*3+2];
	PORTC  = displayBuffer[scan*3+2];

	//### equalization post-delay 
	if ( on_count < MID_LED ){
		delayMicroseconds( dd[on_count] );
		alloff();
	}

	period = 0;
	msec++;
	if (msec >= 10){
		msec = 0;
		period |= TENMS;
	}
	
	scan++;
	if (scan >= 16) {
		scan = 0;
		period |= FRAME;

		// If the page should be flipped, do it here.
		if (videoFlipPage && displayMode == DOUBLE_BUFFER)
		{
		    // TODO: is this an atomic operation?
		    videoFlipPage = false;

		    byte* temp = displayBuffer;
		    displayBuffer = workBuffer;
		    workBuffer = temp;
		}
	}
}


/* -----------------------------------------------------------------  */
/** Constructor : Initialize the interrupt code. 
 * should be called in setup();
 */
void Charlieplexing::Init(byte mode)
//void iLedSign::Init(byte mode)
{
	//### MsTimer2 setting
	MsTimer2::set(1,scanLED);
	MsTimer2::start();

	//### make using port bit pattern
	for (int j=0; j<16; j++){
		byte k = ledMap[ j ];
		if ( k < 8 ){
			out_d |= _BV( k );      // line 76543210
		} else if ( k <14 ){
		    out_b |= _BV( k - 8 );  // line xxDCBA98
		} else {
		    out_c |= _BV( k - 14 ); // line xxJIHGFE
		}
	}
	
	//### make using input port bit pattern
	for (int j=12; j<16; j++){  // LA5,LA6,LA7,LA8
		byte k = ledMap[ j ];
		if ( k < 8 ){
			in_d |= _BV( k );      // line 76543210
		} else if ( k <14 ){
		    in_b |= _BV( k - 8 );  // line xxDCBA98
		} else {
		    in_c |= _BV( k - 14 ); // line xxJIHGFE
		}
	}
#ifndef ikkeiplexing
#ifdef SQUARE8
	in_c |= 0x01; // A0 pullup
#else
	in_c |= 0x30; // A4,A5 pullup
#endif
#endif

	// Record whether we are in single or double buffer mode
	displayMode = mode;

	// Point the display buffer to the first physical buffer
	displayBuffer = leds[0];

	// If we are in single buffered mode, point the work buffer
	// at the same physical buffer as the display buffer.  Otherwise,
	// point it at the second physical buffer.
	if( displayMode == SINGLE_BUFFER ) {
	    workBuffer = displayBuffer;
	}
	else {
	    workBuffer = leds[1];
	}

	scan = 0;
	msec = 0;

	// Clear the buffer and display it
	Charlieplexing::Clear(0);
	Charlieplexing::Flip(false);
}


/* -----------------------------------------------------------------  */
/* * Clear the screen completely
 * @param blocking if true : wait for flip before returning, if false :
 *                 return immediately.
 */
void Charlieplexing::Flip(bool blocking)
{
    if (displayMode == DOUBLE_BUFFER)
    {
        // Just set the flip flag, the buffer will flip between redraws
        videoFlipPage = true;

        // If we are blocking, sit here until the page flips.
        while (blocking && videoFlipPage) {
            delay(1);
        }
    }
}


/* -----------------------------------------------------------------  */
/** Clear the screen completely
 * @param set if 1 : make all led ON, if not set or 0 : make all led OFF
 */
void Charlieplexing::Clear(int set) {
    for(int x=0;x<16;x++)  
        for(int y=0;y<8;y++) 
            Set(x,y,set);
}

/* -----------------------------------------------------------------  */
/** Shift left the screen completely
 */
void Charlieplexing::ShiftLeft() {
    for(int x=1;x<16;x++)  
        for(int y=0;y<8;y++) 
            Set(x-1,y,Get(x,y));
}

/* -----------------------------------------------------------------  */
/** Shift right the screen completely
 */
void Charlieplexing::ShiftRight() {
    for(int x=1;x<16;x++)  
        for(int y=0;y<8;y++) 
            Set(16-x,y,Get(15-x,y));
}

/* -----------------------------------------------------------------  */
/** Clear an horizontal line completely
 * @param y is the y coordinate of the line to clear/light [0-8]
 * @param set if 1 : make all led ON, if not set or 0 : make all led OFF
 */
void Charlieplexing::Horizontal(int y, int set) {
    for(int x=0;x<16;x++)  
        Set(x,y,set);
}


/* -----------------------------------------------------------------  */
/** Clear a vertical line completely
 * @param x is the x coordinate of the line to clear/light [0-13]
 * @param set if 1 : make all led ON, if not set or 0 : make all led OFF
 */
void Charlieplexing::Vertical(int x, int set) {
    for(int y=0;y<8;y++)  
        Set(x,y,set);
}

/* -----------------------------------------------------------------  */
void Charlieplexing::Vwrite(int x, byte data) {
    for(int y=0; y<8; y++){
    	if( data & (1<<y)){
    		Set(x, y, 1);
      	} else {
			Set(x, y, 0);
      	}
    }
}


/* -----------------------------------------------------------------  */
void Charlieplexing::write(byte y, const uint16_t values[], byte size)
{
	for(byte i=0;i<size;i++){
//		_buffer[(y++) & 0x07] = values[i];
		for(byte j=0;j<16;j++){
			if ((y+i) < 8){
				Set(j  , ((y+i) & 0x07), bitRead( values[i],15-j));
			} else {
				Set(j+8, ((y+i) & 0x07), bitRead( values[i],15-j));
			}
		}
	}
}

/* -----------------------------------------------------------------  */
void Charlieplexing::write90(byte y, const uint16_t values[], byte size)
{
	for(byte i=0;i<size;i++){
//		_buffer[(y++) & 0x07] = values[i];
		for(byte j=0;j<16;j++){
			Set(((y+i) & 0x0F), j,   bitRead( values[i],15-j));
		}
	}
}

/* -----------------------------------------------------------------  */
void Charlieplexing::Set(byte x, byte y, byte c)
{
	byte pin_high;
	if ( x >= 8 ){
		pin_high = ledMap[ y ];
	} else {
		pin_high = ledMap[ y + 8 ];
	}
	if (c == 1) {
		if ( pin_high < 8 ){
		    workBuffer[ x*3     ] |= _BV( pin_high );
		}else if ( pin_high < 14 ){
		    workBuffer[ x*3 + 1 ] |= _BV( pin_high - 8 );
		}else{
		    workBuffer[ x*3 + 2 ] |= _BV( pin_high - 14 );
		}
	} else {
		if ( pin_high < 8 ){
		    workBuffer[ x*3     ] &= ~_BV( pin_high );
		}else if ( pin_high < 14 ){
		    workBuffer[ x*3 + 1 ] &= ~_BV( pin_high - 8 );
		}else{
		    workBuffer[ x*3 + 2 ] &= ~_BV( pin_high - 14 );
		}
	}
}

/* -----------------------------------------------------------------  */
byte Charlieplexing::Get(byte x, byte y)
{
	byte pin_high;
	byte b;
	byte c;
	if ( x >= 8 ){
		pin_high = ledMap[ y ];
	} else {
		pin_high = ledMap[ y + 8 ];
	}

	if ( pin_high < 8 ){
	    b = workBuffer[ x*3     ] & _BV( pin_high );
	}else if ( pin_high < 14 ){
	    b = workBuffer[ x*3 + 1 ] & _BV( pin_high - 8 );
	}else{
	    b = workBuffer[ x*3 + 2 ] & _BV( pin_high - 14 );
	}
	if (b){
		c = 1;
	} else {
		c= 0;
	}
	return c;
}


/* -----------------------------------------------------------------  */
/** GetSW : get switch status
 */
byte Charlieplexing::GetSW(byte edge)
{
	static byte previous_swdata;
	
	byte flag = swdata;
	if (edge){
		flag &= ~previous_swdata;
		previous_swdata = swdata;
	}
 	return (flag);
} 


/* -----------------------------------------------------------------  */
/** GetPeriod : get period flag
 */
byte Charlieplexing::GetPeriod(void)
{
	static byte previous_period;
	byte flag = period & ~previous_period;
	previous_period = period;

 	return (flag);
} 


